from datetime import datetime
import numpy as np
from hestia_earth.schema import NodeType

from hestia_earth.utils.tools import (
    non_empty_value,
    non_empty_list,
    is_term,
    current_time_ms,
    safe_parse_float,
    safe_parse_date,
    list_sum,
    list_average,
    flatten,
    get_dict_key,
    to_precision,
    is_number,
    is_boolean,
    omit,
    pick,
    unique_values,
)


def test_non_empty_value():
    assert not non_empty_value("")
    assert not non_empty_value([])
    assert not non_empty_value({})
    assert non_empty_value("test") is True
    assert non_empty_value(False) is True
    assert non_empty_value(1) is True
    assert non_empty_value(np.float64(10)) is True


def test_non_empty_list():
    assert non_empty_list(["", 1, [], False]) == [1, False]


def test_is_term():
    assert not is_term({"@type": NodeType.CYCLE.value})
    assert is_term({"type": NodeType.TERM.value}) is True


def test_current_time_ms():
    assert current_time_ms() > 1000


def test_safe_parse_float():
    assert safe_parse_float("123.456") == 123.456
    assert safe_parse_float("abcd", 10) == 10
    assert safe_parse_float(np.nan, 1) == 1


def test_safe_parse_date():
    assert safe_parse_date("2020-01-01") == datetime(2020, 1, 1)
    assert safe_parse_date("abcd", datetime(2020, 1, 1)) == datetime(2020, 1, 1)


def test_list_sum():
    assert list_sum([0, 1]) == 1
    assert list_sum([]) == 0
    assert list_sum(None) == 0
    assert list_sum("value") == 0

    assert list_sum([], None) is None
    assert list_sum([None], None) is None

    assert list_sum(["random"], None) is None


def test_list_average():
    assert list_average([0, 1]) == 0.5
    assert list_average([]) == 0
    assert list_average(None) == 0
    assert list_average("value") == 0

    assert list_average([], None) is None
    assert list_average([None], None) is None

    assert list_average(["random"], None) is None


def test_flatten():
    assert flatten([[0, 1], 2, [3, 4]]) == [0, 1, 2, 3, 4]


def test_get_dict_key():
    value = {"a": "test", "b": {"id": "id"}, "c": {"d": [{"id": 1}, {"id": 2}]}}
    assert get_dict_key(value, "a") == "test"
    assert get_dict_key(value, "b.id") == "id"
    assert get_dict_key(value, "c.d.id") == [1, 2]


def test_to_precision():
    assert to_precision(0) == 0

    assert to_precision(0.45249) == 0.452
    assert to_precision(1.45213) == 1.45
    assert to_precision(144.5213) == 145

    assert to_precision(0.45249, 1) == 0.5
    assert to_precision(1.45213, 1) == 1
    assert to_precision(145, 1) == 100

    assert to_precision(0.000152, 3) == 0.000152
    assert to_precision(9089080.000000001111, 3) == 9090000


def test_is_number():
    assert is_number(10) is True
    assert is_number(10.01) is True
    assert not is_number(True)
    assert not is_number(False)
    assert not is_number("test")
    assert not is_number("0")


def test_is_boolean():
    assert not is_boolean(10)
    assert not is_boolean(10.01)
    assert is_boolean(True) is True
    assert is_boolean(False) is True
    assert not is_boolean("test")
    assert not is_boolean("0")


def test_omit():
    assert omit({"a": 1, "b": 2}, ["a"]) == {"b": 2}


def test_pick():
    assert pick({"a": 1, "b": 2}, ["a"]) == {"a": 1}


def test_unique_values():
    values = [{"@id": 1, "name": "a"}, {"@id": 2, "name": "a"}]
    assert unique_values(values) == values

    values = [{"@id": 1, "name": "a"}, {"@id": 1, "name": "b"}]
    assert unique_values(values) == [{"@id": 1, "name": "b"}]



import os

from .tools import non_empty_value


def api_url() -> str:
    return os.getenv("API_URL", "https://api.hestia.earth")


def api_access_token() -> str:
    return os.getenv("API_ACCESS_TOKEN")


def web_url() -> str:
    return os.getenv("WEB_URL", "https://www.hestia.earth")


def join_args(values) -> str:
    return "&".join(list(filter(non_empty_value, values))).strip()


def request_url(base_url: str, **kwargs) -> str:
    args = list(
        map(
            lambda key: (
                "=".join([key, str(kwargs.get(key))]) if kwargs.get(key) else None
            ),
            kwargs.keys(),
        )
    )
    return "?".join(list(filter(non_empty_value, [base_url, join_args(args)]))).strip()



import os

_sns_client = None  # noqa: F824


# improves speed for connecting on subsequent calls
def _get_sns_client():
    global _sns_client
    import boto3

    region_name = os.getenv("AWS_REGION")
    _sns_client = (
        boto3.session.Session().client("sns", region_name=region_name)
        if _sns_client is None
        else _sns_client
    )
    return _sns_client



from os.path import join
import json
import numpy as np

from .tools import current_time_ms, non_empty_list, flatten
from .api import find_related
from .storage._s3_client import (
    _load_from_bucket,
    _upload_to_bucket,
    _last_modified,
    _read_metadata,
    _update_metadata,
    _exists_in_bucket,
)
from .storage._sns_client import _get_sns_client

PROGRESS_EXT = ".progress"
CALC_FOLDER = "recalculated"
METADATA_STAGE_KEY = "stage"
METADATA_PROGRESS_KEY = "calculating"


# fix error "Object of type int64 is not JSON serializable"
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


def to_string(data: dict, indent: int = None):
    return json.dumps(data, indent=indent, ensure_ascii=False, cls=NpEncoder)


def to_bytes(data: dict):
    return to_string(data).encode("utf8")


def upload_json(bucket_name: str, file_key: str, body: dict):
    return _upload_to_bucket(
        bucket=bucket_name,
        key=file_key,
        body=to_bytes(body),
        content_type="application/json",
    )


def _to_file_progress(filepath: str):
    return (
        filepath.replace(".csv", PROGRESS_EXT)
        .replace(".json", PROGRESS_EXT)
        .replace(".hestia", PROGRESS_EXT)
    )


def handle_result(
    bucket_name: str, file_key: str, step: str, start: int, content: dict
):
    filepath = _to_file_progress(file_key)

    # try to read existing progress to update the time per step
    try:
        data = json.loads(_load_from_bucket(bucket_name, filepath))
    except Exception:
        data = {}

    return upload_json(
        bucket_name,
        filepath,
        {
            **data,
            "step": step,
            "time": {
                **(
                    data.get("time", {})
                    if isinstance(data.get("time", {}), dict)
                    else {}
                ),
                step: current_time_ms() - start,
            },
            **content,
        },
    )


def handle_error(
    bucket_name: str,
    file_key: str,
    step: str,
    start: int,
    err: str = "",
    stack: str = "",
    errors=[],
    warnings=[],
    extras: dict = {},
):
    return handle_result(
        bucket_name,
        file_key,
        step,
        start,
        extras
        | {
            "success": False,
            "error": {
                "message": err,
                "stack": stack,
                "errors": errors,
                "warnings": warnings,
            },
            "warning": {"warnings": warnings},
        },
    )


def handle_success(
    bucket_name: str, file_key: str, step: str, start: int, extras: dict = {}
):
    return handle_result(bucket_name, file_key, step, start, extras | {"success": True})


def publish_result(
    topic_arn: str,
    bucket_name: str,
    file_key: str,
    filepath: str,
    step: str,
    success: bool,
):
    return _get_sns_client().publish(
        TopicArn=topic_arn,
        Message=to_string(
            {"bucket": bucket_name, "key": file_key, "filepath": filepath}
        ),
        MessageAttributes={
            "functionName": {
                "DataType": "String",
                "StringValue": step + ("Done" if success else "Error"),
            }
        },
    )


def _parse_event_s3(event: dict):
    return {
        "bucket": event["s3"]["bucket"]["name"],
        "key": event["s3"]["object"]["key"],
    }


def _parse_event_SNS(event: dict):
    event = event.get("Sns", {})
    data = json.loads(event.get("Message", "{}"))
    attributes: dict = event.get("MessageAttributes", {})
    data["attributes"] = {key: value.get("Value") for key, value in attributes.items()}
    return data


def _parse_event_SQS(event: dict):
    condition = event.get("requestContext", {}).get("condition")
    return (
        _get_data_from_event(event.get("requestPayload", {}))
        if condition != "RetriesExhausted"
        else None
    )


def _get_data_from_event(event):  # noqa: C901
    if isinstance(event, dict):
        if "s3" in event:
            return _parse_event_s3(event)
        # invoked when running asynchronously
        if "Sns" in event:
            return _parse_event_SNS(event)
        # invoked through http event
        if "body" in event:
            return _get_data_from_event(json.loads(event.get("body", "{}")))
        # invoked through s3 put object
        if "Records" in event:
            return flatten(map(_get_data_from_event, event.get("Records", [])))
        # invoked when calculation timedout or failed
        if "requestPayload" in event:
            return _parse_event_SQS(event)
        return event
    if isinstance(event, str):
        return _get_data_from_event(json.loads(event))


def parse_event(event: dict):
    data = _get_data_from_event(event)
    return non_empty_list(flatten(data) if isinstance(data, list) else [data])


def _node_type(node: dict):
    return node.get("@type", node.get("type"))


def _node_id(node: dict):
    return node.get("@id", node.get("id"))


def _node_path(node: dict, folder: str = ""):
    return join(folder, _node_type(node), f"{_node_id(node)}.jsonld")


def _load_node(bucket: str, file_key: str):
    return json.loads(_load_from_bucket(bucket, file_key))


def _cache_path(node: dict):
    return join(_node_type(node), f"{_node_id(node)}.cache")


def _has_cache(bucket: str, node: dict):
    return _exists_in_bucket(bucket, _cache_path(node))


def is_calculating(bucket: str, node: dict, folder: str = ""):
    return (
        _read_metadata(bucket, _node_path(node, folder)).get(
            METADATA_PROGRESS_KEY, "false"
        )
        == "true"
    )


def set_calculating(bucket: str, node: dict, in_progress: bool, folder: str = ""):
    return _update_metadata(
        bucket,
        _node_path(node, folder),
        {METADATA_PROGRESS_KEY: str(in_progress).lower()},
    )


def get_stage(bucket: str, node: dict, folder: str = CALC_FOLDER):
    stage = _read_metadata(bucket, _node_path(node, folder=CALC_FOLDER)).get(
        METADATA_STAGE_KEY
    )
    return int(stage) if stage else stage


def load_cache(bucket: str, node: dict):
    """
    Return the cache data for the node.

    Parameters
    ----------
    bucket : str
        The bucket where the cache is stored.
    node : dict
        The Node which is connected to other nodes (source).

    Returns
    -------
    dict
        The cached data.
    """
    cache_path = join(node["@type"], f"{node['@id']}.cache")
    try:
        return json.loads(_load_from_bucket(bucket, cache_path))
    except Exception:
        print("No cache found for", cache_path)
        return {}


def _filter_by_type(nodes: list, type: str):
    return [n for n in nodes if n.get("@type", n.get("type")) == type]


def _find_related_nodes(
    from_type: str, from_id: str, related_type: str, related_key: str
):
    should_find_related = related_key == "related"
    print("Find related nodes from API", from_type, from_id, related_key, related_type)
    return (
        find_related(from_type, from_id, related_type, limit=10000)
        if should_find_related
        else []
    )


def _get_cached_nodes(
    cache: dict, related_key: str, from_type: str, from_id: str, to_type: str
):
    # if key is in cache, use nodes in cache, otherwise use API
    if related_key in cache:
        nodes = _filter_by_type(cache.get(related_key, []), to_type)
        print("Using cached data to", related_key, to_type, nodes)
        return list(
            map(
                lambda node: {"@type": to_type, "@id": node.get("@id", node.get("id"))},
                nodes,
            )
        )
    else:
        return _find_related_nodes(from_type, from_id, to_type, related_key)


def get_related_nodes(node: dict, related_key: str, related_type: str, cache: dict):
    """
    Given a node, return all related nodes.

    Parameters
    ----------
    node : dict
        The Node which is connected to other nodes (source).
    related_key : str
        Either `nested` or `related`.
    related_type : str
        Related node `@type`.
    cache : dict
        Cache data of the source Node. Can contain `related` and `nested` nodes.
        When no provided, only `related` nodes can be found.

    Returns
    -------
    List[dict]
        The related nodes.
    """
    from_type = node.get("@type", node.get("type"))
    from_id = node.get("@id", node.get("id"))

    related_nodes = (
        _get_cached_nodes(cache or {}, related_key, from_type, from_id, related_type)
        or []
    )

    return list(
        {f"{node['@type']}/{node['@id']}": node for node in related_nodes}.values()
    )


def get_related_nodes_data(
    bucket_name: str, node: dict, related_key: str, related_type: str, cache: dict
):
    """
    Given a node, return all related nodes with extra data.

    Parameters
    ----------
    bucket_name : str
        The name of the bucket where the nodes are stored.
    node : dict
        The Node which is connected to other nodes (source).
    related_key : str
        Either `nested` or `related`.
    related_type : str
        Related node `@type`.
    cache : dict
        Cache data of the source Node. Can contain `related` and `nested` nodes.
        When no provided, only `related` nodes can be found.

    Returns
    -------
    List[dict]
        The related nodes with extra data: `indexed_at`, `recalculated_at` and `recalculated_stage`.
    """
    related_nodes = get_related_nodes(
        node=node, related_key=related_key, related_type=related_type, cache=cache
    )

    return [
        node
        | {
            "indexed_at": _last_modified(bucket=bucket_name, key=_node_path(node)),
            "recalculated_at": _last_modified(
                bucket=bucket_name, key=_node_path(node, folder=CALC_FOLDER)
            ),
            "recalculated_stage": get_stage(bucket_name, node),
            "is_calculating": is_calculating(bucket_name, node),
        }
        for node in related_nodes
    ]



import time
from dateutil.parser import parse
from statistics import mean
from functools import reduce
from math import log10, floor
import numpy
from hestia_earth.schema import NodeType


def to_precision(number: float, digits: int = 3) -> float:
    return (
        0 if not number else round(number, digits - int(floor(log10(abs(number)))) - 1)
    )


def non_empty_value(value) -> bool:
    """
    Check if a value is empty.

    Parameters
    ----------
    value
        Either a string, a list, a number or None.

    Returns
    -------
    bool
        True if the value is not en empty string or an empty list.
    """
    return all(
        [
            value != "",
            value is not None,
            not isinstance(value, list) or value != [],
            not isinstance(value, dict) or value != {},
        ]
    )


def non_empty_list(values: list) -> list:
    """
    Filter list removing empty values.

    Parameters
    ----------
    values
        A list of values.

    Returns
    -------
    list
        List without empty values.
    """
    return list(filter(non_empty_value, values))


def is_node_of(node_type) -> bool:
    """
    Check wether node is of a certain HESTIA Type.

    Parameters
    ----------
    node_type
        The type to check for.
    node
        The node.

    Returns
    -------
    bool
        True if matches type.
    """
    return (
        lambda node: isinstance(node, dict)
        and node.get("type", node.get("@type")) == node_type.value
    )


def is_term(node: dict) -> bool:
    """
    Check wether node is a `Term`.

    Parameters
    ----------
    node
        The node.

    Returns
    -------
    bool
        True if it is a `Term`.
    """
    return is_node_of(NodeType.TERM)(node)


def current_time_ms():
    """
    Get the time in ms since EPOCH.

    Returns
    -------
    int
        Time in milliseconds.
    """
    return float(time.time() * 1000)


def safe_parse_float(value: str, default=0):
    """
    Parse a string into a float.

    Parameters
    ----------
    value
        The string value to parse.
    default
        The default value if parsing not possible.

    Returns
    -------
    float
        The value as float or default value.
    """
    try:
        value = float(value)
        return default if numpy.isnan(value) else value
    except Exception:
        return default


def safe_parse_date(date=None, default=None):
    """
    Parse a string into a date.

    Parameters
    ----------
    value
        The string value to parse.
    default
        The default value if parsing not possible.

    Returns
    -------
    datetime
        The value as datetime or default value.
    """
    try:
        return parse(str(date), fuzzy=True)
    except Exception:
        return default


def is_number(value):
    """
    Return `True` if the value is either an `int` or a `float`.
    """
    return all(
        [
            not isinstance(value, bool),  # True is apparently considered an `int`
            any([isinstance(value, int), isinstance(value, float)]),
        ]
    )


def is_boolean(v):
    """
    Return `True` if the value is a `bool`.
    """
    return isinstance(v, bool)


def list_average(value: list, default=0):
    """
    Returns the average over a list of numbers.

    Parameters
    ----------
    value
        A list of numbers.
    default
        The default value if the value does not contain a list of numbers.

    Returns
    -------
    float
        The average of the values.
    """
    values = (
        non_empty_list(value)
        if value and isinstance(value, list) and all(map(is_number, value))
        else []
    )
    return mean(values) if values else default


def list_sum(value: list, default=0):
    """
    Returns the sum over a list of numbers.

    Parameters
    ----------
    value
        A list of numbers.
    default
        The default value if the value does not contain a list of numbers.

    Returns
    -------
    float
        The sum of the values.
    """
    values = (
        non_empty_list(value)
        if value and isinstance(value, list) and all(map(is_number, value))
        else []
    )
    return sum(values) if values else default


def flatten(values: list):
    """
    Flattens a two-dimensional list into a one-dimensional list.

    Parameters
    ----------
    values
        A list of list.

    Returns
    -------
    list
        A list of single values.
    """
    return list(
        reduce(lambda x, y: x + (y if isinstance(y, list) else [y]), values, [])
    )


def _get_by_key(x, y):
    return (
        x
        if x is None
        else (
            x.get(y)
            if isinstance(x, dict)
            else list(map(lambda v: get_dict_key(v, y), x))
        )
    )


def get_dict_key(value: dict, key: str):
    return reduce(lambda x, y: _get_by_key(x, y), key.split("."), value)


def omit(values: dict, keys: list) -> dict:
    return {k: v for k, v in values.items() if k not in keys}


def pick(value: dict, keys: list) -> dict:
    return {k: v for k, v in value.items() if k in keys}


def unique_values(values: list, key: str = "@id"):
    return list({v[key]: v for v in values}.values())



from functools import reduce
from typing import Any
import requests
from io import StringIO
import pandas as pd

from .storage import _load_from_storage
from .request import request_url, web_url

_GLOSSARY_FOLDER = "glossary/lookups"
_memory = {}


def _memory_wrapper(key: str, func):
    global _memory  # noqa: F824
    _memory[key] = _memory[key] if key in _memory else func()
    return _memory[key]


def _read_csv(value: str) -> pd.DataFrame:
    return pd.read_csv(value, na_values=["-", ""])


def _read_csv_from_string(data: str) -> pd.DataFrame:
    return _read_csv(StringIO(data))


def is_missing_value(value):
    return pd.isna(value) or value is None or value == "" or value == "-"


def load_lookup(filepath: str, keep_in_memory: bool = False):
    """
    Import local lookup table as csv file into a `numpy.recarray`.

    Parameters
    ----------
    filepath : str
        The path of csv file on the local file system.
    keep_in_memory: bool
        Set to True if you want to store the file in memory for later use.

    Returns
    -------
    numpy.recarray
        The `numpy.recarray` converted from the csv content.
    """

    def load():
        return _read_csv(filepath)

    return _memory_wrapper(filepath, load) if keep_in_memory else load()


def _download_lookup_data(filename: str):
    filepath = f"{_GLOSSARY_FOLDER}/{filename}"

    def fallback():
        url = request_url(f"{web_url()}/{filepath}")
        data = requests.get(url).content.decode("utf-8")
        return data if data and "<html" not in data else None

    try:
        data = _load_from_storage(filepath, glossary=True)
        return data.decode("utf-8") if data else None
    except ImportError:
        return fallback()


def download_lookup(filename: str, keep_in_memory: bool = True):
    """
    Download lookup table from HESTIA as csv into a `numpy.recarray`.

    Parameters
    ----------
    filename : str
        The name on the file on the HESTIA lookup repository.
    keep_in_memory: bool
        Set to False if you do NOT want to store the file in memory for later use.
    build_index : bool
        Set to False to skip trying to build an index.

    Returns
    -------
    numpy.recarray
        The `numpy.recarray` converted from the csv content.
    """

    def load():
        data = _download_lookup_data(filename)
        return _read_csv_from_string(data) if data else None

    try:
        return _memory_wrapper(filename, load) if keep_in_memory else load()
    except Exception:
        return None


def column_name(key: str):
    """
    Deprecated. Columns are no longer renamed.
    """
    return key


def _parse_value(value: str):
    """Automatically converts the value to float or bool if possible"""
    try:
        return (
            True
            if str(value).lower() == "true"
            else False if str(value).lower() == "false" else float(value)
        )
    except Exception:
        return value


def _get_single_table_value(df: pd.DataFrame, col_match: str, col_match_with, col_val):
    filtered_df = df[df[col_match] == col_match_with]
    return None if filtered_df.empty else filtered_df[col_val].iloc[0]


def get_table_value(
    lookup: pd.DataFrame,
    col_match: str,
    col_match_with: str,
    col_val: Any,
    default_value="",
):
    """
    Get a value matched by one or more columns from a `numpy.recarray`.

    Parameters
    ----------
    lookup : DataFrame
        The value returned by the `download_lookup` function.
    col_match : str
        Which `column` should be used to find data in. This will restrict the rows to search for.
        Can be a single `str` or a list of `str`. If a list is used, must be the same length as `col_match_with`.
    col_match_with: str
        Which column `value` should be used to find data in. This will restrict the rows to search for.
    col_val: str
        The column which contains the value to look for.
    default_value : Any
        A value to return when none if found in the data.

    Returns
    -------
    str
        The value found or `None` if no match.
    """
    try:
        value = _get_single_table_value(lookup, col_match, col_match_with, col_val)
        return default_value if is_missing_value(value) else _parse_value(value)
    except Exception:
        return None


def find_term_ids_by(lookup: pd.DataFrame, col_match: str, col_match_with: str):
    """
    Find `term.id` values where a column matches a specific value.

    Parameters
    ----------
    lookup : DataFrame
        The value returned by the `download_lookup` function.
    col_match : str
        Which `column` should be used to find data in. This will restrict the rows to search for.
        Can be a single `str` or a list of `str`. If a list is used, must be the same length as `col_match_with`.
    col_match_with: str
        Which column `value` should be used to find data in. This will restrict the rows to search for.

    Returns
    -------
    list[str]
        The list of `term.id` that matched the expected column value.
    """
    filtered_df = lookup[lookup[col_match] == col_match_with]
    term_ids = (
        filtered_df["term.id"].unique().tolist()
        if "term.id" in filtered_df.columns
        else []
    )
    return list(map(str, term_ids))


def extract_grouped_data(data: str, key: str) -> str:
    """
    Extract value from a grouped data in a lookup table.

    Example:
    - with data: `Average_price_per_tonne:106950.5556;1991:-;1992:-`
    - get the value for `Average_price_per_tonne` = `106950.5556`

    Parameters
    ----------
    data
        The data to parse. Must be a string in the format `<key1>:<value>;<key2>:<value>`
    key
        The key to extract the data. If not present, `None` will be returned.

    Returns
    -------
    str
        The value found or `None` if no match.
    """
    grouped_data = (
        reduce(
            lambda prev, curr: {**prev, **{curr.split(":")[0]: curr.split(":")[1]}},
            data.split(";"),
            {},
        )
        if data is not None and isinstance(data, str) and len(data) > 1
        else {}
    )
    value = grouped_data.get(key)
    return None if is_missing_value(value) else _parse_value(value)


def extract_grouped_data_closest_date(data: str, year: int) -> str:
    """
    Extract date value from a grouped data in a lookup table.

    Example:
    - with data: `2000:-;2001:0.1;2002:0;2003:0;2004:0;2005:0`
    - get the value for `2001` = `0.1`

    Parameters
    ----------
    data
        The data to parse. Must be a string in the format `<key1>:<value>;<key2>:<value>`
    year
        The year to extract the data. If not present, the closest date data will be returned.

    Returns
    -------
    str
        The closest value found.
    """
    data_by_date = (
        reduce(
            lambda prev, curr: (
                {**prev, **{curr.split(":")[0]: curr.split(":")[1]}}
                if len(curr) > 0 and not is_missing_value(curr.split(":")[1])
                else prev
            ),
            data.split(";"),
            {},
        )
        if data is not None and isinstance(data, str) and len(data) > 1
        else {}
    )
    dist_years = list(data_by_date.keys())
    closest_year = (
        min(dist_years, key=lambda x: abs(int(x) - year))
        if len(dist_years) > 0
        else None
    )
    return (
        None if closest_year is None else _parse_value(data_by_date.get(closest_year))
    )


def lookup_term_ids(lookup: pd.DataFrame):
    """
    Get the `term.id` values from a lookup.

    Parameters
    ----------
    lookup : DataFrame
        The value returned by the `download_lookup` function.

    Returns
    -------
    list[str]
        The `term.id` values from the lookup.
    """
    return (
        list(map(str, lookup["term.id"].tolist()))
        if "term.id" in lookup.columns
        else []
    )


def lookup_columns(lookup: pd.DataFrame):
    """
    Get the columns from a lookup.

    Parameters
    ----------
    lookup : DataFrame
        The value returned by the `download_lookup` function.

    Returns
    -------
    list[str]
        The columns from the lookup.
    """
    return list(lookup.columns)



from hestia_earth.utils.date import diff_in_years, is_in_days, is_in_months


def test_diff_in_years():
    assert diff_in_years("1990-01-01", "1999-02-01") == 9.1


def test_is_in_days():
    assert not is_in_days("2000")
    assert not is_in_days("2000-01")
    assert is_in_days("2000-01-01")


def test_is_in_months():
    assert not is_in_months("2000")
    assert is_in_months("2000-01")
    assert not is_in_months("2000-01-01")



import pytest
import pandas as pd

from .utils import fixtures_path
from hestia_earth.utils.lookup import (
    load_lookup,
    get_table_value,
    find_term_ids_by,
    download_lookup,
    extract_grouped_data,
    extract_grouped_data_closest_date,
    _get_single_table_value,
    lookup_term_ids,
    lookup_columns,
)


def test_load_lookup_type():
    lookup = load_lookup(f"{fixtures_path}/lookup.csv")
    assert isinstance(lookup, pd.DataFrame)


@pytest.mark.parametrize(
    "col_match,col_match_with,col_val,expected",
    [
        ("Col1", "val10", "Col3", "val30"),
        ("Col1", "val10", "Col5", None),
        ("Col1", "val10", "Col4", ""),
    ],
)
def test_get_table_value(col_match, col_match_with, col_val, expected):
    lookup = load_lookup(f"{fixtures_path}/lookup.csv")
    assert get_table_value(lookup, col_match, col_match_with, col_val) == expected


def test_get_table_value_no_lookup():
    assert not get_table_value(None, "Col10", "val10", "Col3")


def test_get_table_value_default_value():
    lookup = load_lookup(f"{fixtures_path}/lookup.csv")
    assert get_table_value(lookup, "Col2", "val22", "Col1") == ""

    lookup = download_lookup("crop.csv")
    assert (
        get_table_value(lookup, "term.id", "genericCropSeed", "Plantation_density")
        == ""
    )
    assert (
        get_table_value(
            lookup, "term.id", "fixedNitrogen", "Combustion_Factor_crop_residue"
        )
        == ""
    )


def test_find_term_ids_by():
    lookup = download_lookup("crop.csv")
    assert "wheatGrain" in find_term_ids_by(
        lookup, "cropGroupingFAO", "Temporary crops"
    )


def test_handle_missing_float_value():
    filename = "measurement.csv"
    lookup = download_lookup(filename)
    assert get_table_value(lookup, "term.id", "rainfallPeriod", "maximum") == ""


def test_handle_missing_lookup_value():
    filename = "region-crop-cropGroupingFaostatProduction-price.csv"
    lookup = download_lookup(filename)
    assert get_table_value(lookup, "term.id", "GADM-CYP", "Sugar crops nes") is None


def test_extract_grouped_data_no_data():
    assert not extract_grouped_data("", "2000")
    assert not extract_grouped_data("-", "2000")


def test_extract_grouped_data():
    data = "Average_price_per_tonne:106950.5556;1991:-;1992:-"
    assert extract_grouped_data(data, "Average_price_per_tonne") == 106950.5556
    assert extract_grouped_data(data, "2010") is None


def test_extract_grouped_data_lookup():
    filename = "region-crop-cropGroupingFaostatProduction-price.csv"
    lookup = download_lookup(filename)
    data = get_table_value(lookup, "term.id", "GADM-NPL", "Chick peas, dry")
    assert extract_grouped_data(data, "2000") is None
    assert extract_grouped_data(data, "2012") is not None

    filename = "region-animalProduct-animalProductGroupingFAO-price.csv"
    lookup = download_lookup(filename)
    data = get_table_value(
        lookup, "term.id", "GADM-NPL", "Eggs from other birds in shell, fresh, n.e.c."
    )
    assert extract_grouped_data(data, "2000") is None
    assert extract_grouped_data(data, "2012") is not None


def test_get_single_table_value_float_values():
    filename = "ecoClimateZone.csv"
    lookup = download_lookup(filename)
    column = "STEHFEST_BOUWMAN_2006_N2O-N_FACTOR"
    assert _get_single_table_value(lookup, "ecoClimateZone", 11, column) == -0.3022


def test_extract_grouped_data_closest_date_no_data():
    assert not extract_grouped_data_closest_date("", 2000)
    assert not extract_grouped_data_closest_date("-", 2000)


def test_extract_grouped_data_closest_date():
    data = "2000:-;2001:0.1;2002:0.2;2003:0.3;2004:0.4;2005:0.5"
    assert extract_grouped_data_closest_date(data, 2000) == 0.1
    assert extract_grouped_data_closest_date(data, 2001) == 0.1
    assert extract_grouped_data_closest_date(data, 2020) == 0.5


def test_lookup_term_ids():
    assert "wheatGrain" in lookup_term_ids(download_lookup("crop.csv"))


def test_lookup_columns():
    assert "term.id" in lookup_columns(download_lookup("crop.csv"))


def test_get_data_advanced():
    lookup = download_lookup("liveAnimal.csv")
    value = get_table_value(
        lookup, "term.id", "sheepRam", "ratioCPregnancyNetEnergyPregnancyIpcc2019"
    )
    assert value == ""


def test_grouping_with_comma():
    lookup = download_lookup("animalProduct.csv")
    term_id = "meatChickenReadyToCookWeight"
    value = get_table_value(lookup, "term.id", term_id, "animalProductGroupingFAO")
    assert value == "Meat of chickens, fresh or chilled"



from hestia_earth.schema import TermTermType

from hestia_earth.utils.model import (
    find_term_match,
    find_primary_product,
    filter_list_term_type,
    convert_value,
    linked_node,
)


def test_find_term_match():
    values = [
        {"term": {"@id": "my term"}, "methodModel": {"@id": "model-1"}},
        {"term": {"@id": "another id"}},
        {"term": {"@id": "my term"}, "methodModel": {"@id": "model-2"}},
    ]
    assert find_term_match(values, "my term") == values[0]
    assert (
        find_term_match(values, "my term", match_params={"methodModel.@id": "model-2"})
        == values[2]
    )


def test_find_primary_product():
    primary = {"@type": "Product", "primary": True}
    cycle = {"products": [primary, {"@type": "Product"}]}
    assert find_primary_product(cycle) == primary


def test_filter_list_term_type():
    values = [
        {"term": {"termType": TermTermType.CROP.value}},
        {"term": {"termType": TermTermType.FUEL.value}},
    ]
    assert len(filter_list_term_type(values, TermTermType.CROP)) == 1
    assert (
        len(filter_list_term_type(values, [TermTermType.CROP, TermTermType.FUEL])) == 2
    )


def test_convert_value():
    assert convert_value(1, "m3", "kg", density=553) == 553
    assert convert_value(1, "m3", "L") == 1000
    assert convert_value(553, "kg", "m3", density=553) == 1
    assert convert_value(553, "kg", "L", density=553) == 1000
    assert convert_value(1000, "L", "m3") == 1
    assert round(convert_value(100, "MJ", "kWh")) == 28
    assert convert_value(10, "kWh", "MJ") == 36


def test_linked_node():
    node = {"@type": "Term", "@id": "term-id", "value": 10, "name": "term"}
    assert linked_node(node) == {"@type": "Term", "@id": "term-id", "name": "term"}

    node["@type"] = "Cycle"
    assert linked_node(node) == {"@type": "Cycle", "@id": "term-id"}



import os
import json
import pytest
from unittest.mock import Mock, patch, call

from .utils import fixtures_path
from hestia_earth.utils.tools import flatten
from hestia_earth.utils.pipeline import (
    METADATA_PROGRESS_KEY,
    METADATA_STAGE_KEY,
    parse_event,
    get_related_nodes,
    get_related_nodes_data,
    is_calculating,
    set_calculating,
    get_stage,
    to_string,
)

class_path = "hestia_earth.utils.pipeline"
fixtures_folder = os.path.join(fixtures_path, "pipeline")


def test_parse_event_direct():
    with open(os.path.join(fixtures_folder, "direct-invokation.json")) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [
        {"bucket": "hestia-data-dev", "key": "recalculated/Cycle/1.jsonld"}
    ]


def test_parse_event_sqs():
    with open(os.path.join(fixtures_folder, "sqs.json")) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [
        {
            "bucket": "hestia-data-staging",
            "key": "ImpactAssessment/d3wqcuk_o-vz.jsonld",
            "attributes": {"id": "d3wqcuk_o-vz", "type": "ImpactAssessment"},
        }
    ]


def test_parse_event_sqs_exhausted():
    with open(os.path.join(fixtures_folder, "sqs-exhausted.json")) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == []


def test_parse_event_sns():
    with open(os.path.join(fixtures_folder, "sns.json")) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [
        {
            "bucket": "hestia-data-dev",
            "key": "recalculated/Cycle/1.jsonld",
            "attributes": {"Test": "TestString", "TestBinary": "TestBinary"},
        }
    ]


def test_parse_event_s3():
    with open(os.path.join(fixtures_folder, "s3.json")) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [
        {"bucket": "hestia-data-dev", "key": "recalculated/Cycle/1.jsonld"}
    ]


_CACHE = {
    "nested": [
        {"@type": "Site", "@id": "nested-site-1"},
        {"@type": "Cycle", "@id": "nested-cycle-1"},
        {"@type": "ImpactAssessment", "@id": "nested-ia-1"},
    ],
    "related": [
        {"@type": "Site", "@id": "related-site-1"},
        {"@type": "Cycle", "@id": "related-cycle-1"},
        {"@type": "ImpactAssessment", "@id": "related-ia-1"},
    ],
}


@pytest.mark.parametrize(
    "test_name,node_type,related_key,related_type,cache,expected_nodes",
    [
        (
            "nested-with-cache",
            "Site",
            "nested",
            "Cycle",
            _CACHE,
            [{"@type": "Cycle", "@id": "nested-cycle-1"}],
        ),
        (
            "related-with-cache",
            "Site",
            "related",
            "Site",
            _CACHE,
            [{"@type": "Site", "@id": "related-site-1"}],
        ),
        ("related-empty-cache", "Site", "related", "Cycle", {"related": []}, []),
        ("related-no-cache", "Site", "related", "Cycle", None, _CACHE["related"]),
    ],
)
@patch(f"{class_path}.find_related", return_value=_CACHE["related"])
def test_get_related_nodes(
    mock_find_related: Mock,
    test_name: str,
    node_type: str,
    related_key: str,
    related_type: str,
    cache: dict,
    expected_nodes: list,
):
    nodes = get_related_nodes(
        node={"@type": node_type},
        related_key=related_key,
        related_type=related_type,
        cache=cache,
    )
    assert nodes == expected_nodes, test_name


@pytest.mark.parametrize(
    "test_name,node_type,related_key,related_type,cache,expected_nodes",
    [
        (
            "nested-with-cache",
            "Site",
            "nested",
            "Cycle",
            _CACHE,
            [{"@type": "Cycle", "@id": "nested-cycle-1"}],
        ),
        (
            "related-with-cache",
            "Site",
            "related",
            "Site",
            _CACHE,
            [{"@type": "Site", "@id": "related-site-1"}],
        ),
        ("related-no-cache", "Site", "related", "Cycle", None, _CACHE["related"]),
    ],
)
@patch(f"{class_path}.find_related", return_value=_CACHE["related"])
@patch(f"{class_path}._last_modified", return_value="")
@patch(f"{class_path}.is_calculating", return_value=False)
@patch(f"{class_path}.get_stage", return_value=1)
def test_get_related_nodes_data(
    mock_get_stage: Mock,
    mock_is_calculating: Mock,
    mock_last_modified: Mock,
    mock_find_related: Mock,
    test_name: str,
    node_type: str,
    related_key: str,
    related_type: str,
    cache: dict,
    expected_nodes: list,
):
    bucket = "bucket"
    nodes = get_related_nodes_data(
        bucket_name=bucket,
        node={"@type": node_type},
        related_key=related_key,
        related_type=related_type,
        cache=cache,
    )

    assert nodes == [
        node
        | {
            "indexed_at": "",
            "recalculated_at": "",
            "recalculated_stage": 1,
            "is_calculating": False,
        }
        for node in expected_nodes
    ], test_name

    assert mock_get_stage.call_args_list == [
        call("bucket", node) for node in expected_nodes
    ], test_name

    assert mock_is_calculating.call_args_list == [
        call("bucket", node) for node in expected_nodes
    ], test_name

    assert mock_last_modified.call_args_list == flatten(
        [
            [
                call(bucket="bucket", key=f"{node['@type']}/{node['@id']}.jsonld"),
                call(
                    bucket="bucket",
                    key=f"recalculated/{node['@type']}/{node['@id']}.jsonld",
                ),
            ]
            for node in expected_nodes
        ]
    ), test_name


@patch(f"{class_path}._read_metadata")
def test_get_stage(mock_read_metadata: Mock):
    bucket = "bucket"
    node = {"@type": "Cycle", "@id": "cycle"}

    mock_read_metadata.return_value = {METADATA_STAGE_KEY: "1"}
    assert get_stage(bucket, node) == 1

    mock_read_metadata.return_value = {}
    assert get_stage(bucket, node) is None


@patch(f"{class_path}._read_metadata")
def test_is_calculating(mock_read_metadata: Mock):
    bucket = "bucket"
    node = {"@type": "Cycle", "@id": "cycle"}

    mock_read_metadata.return_value = {METADATA_PROGRESS_KEY: "true"}
    assert is_calculating(bucket, node) is True

    mock_read_metadata.return_value = {METADATA_PROGRESS_KEY: "false"}
    assert not is_calculating(bucket, node)


@patch(f"{class_path}._update_metadata")
def test_set_calculating(mock_update_metadata: Mock):
    bucket = "bucket"
    node = {"@type": "Cycle", "@id": "cycle"}
    set_calculating(bucket, node, True)
    assert mock_update_metadata.call_args_list == [
        call("bucket", "Cycle/cycle.jsonld", {METADATA_PROGRESS_KEY: "true"})
    ]


@pytest.mark.parametrize(
    "data,indent,expected",
    [({"@id": 2}, None, '{"@id": 2}'), ({"@id": 2}, 2, '{\n  "@id": 2\n}')],
)
def test_to_string(data: dict, indent, expected: str):
    assert to_string(data, indent=indent) == expected, expected



from typing import List
from hestia_earth.schema import TermTermType

from .lookup import download_lookup, lookup_term_ids
from .lookup_utils import (
    is_in_system_boundary,
    is_siteType_allowed,
    is_site_measurement_id_allowed,
    is_product_termType_allowed,
    is_product_id_allowed,
    is_input_termType_allowed,
    is_input_id_allowed,
    is_practice_termType_allowed,
    is_practice_id_allowed,
    is_transformation_termType_allowed,
    is_transformation_id_allowed,
)


def emissions_in_system_boundary(
    termType: TermTermType = TermTermType.EMISSION,
) -> List[str]:
    """
    Get all emissions included in HESTIA system boundary.

    Returns
    -------
    List[str]
        List of emission IDs
    """
    lookup = download_lookup(f"{termType.value}.csv")
    # find all emissions in system boundary
    return list(filter(is_in_system_boundary, lookup_term_ids(lookup)))


def cycle_emission_is_in_system_boundary(cycle: dict):
    def filter_term(term_id: str):
        return is_in_system_boundary(term_id) and all(
            map(
                lambda check: check(cycle, term_id),
                [
                    is_siteType_allowed,
                    is_site_measurement_id_allowed,
                    is_product_termType_allowed,
                    is_product_id_allowed,
                    is_input_termType_allowed,
                    is_input_id_allowed,
                    is_practice_termType_allowed,
                    is_practice_id_allowed,
                    is_transformation_termType_allowed,
                    is_transformation_id_allowed,
                ],
            )
        )

    return filter_term


def cycle_emissions_in_system_boundary(
    cycle: dict, termType: TermTermType = TermTermType.EMISSION
):
    """
    Get all emissions relevant for the Cycle, included in HESTIA system boundary.

    Returns
    -------
    List[str]
        List of emission IDs
    """
    lookup = download_lookup(f"{termType.value}.csv")
    # find all emissions in system boundary
    return list(
        filter(cycle_emission_is_in_system_boundary(cycle), lookup_term_ids(lookup))
    )



import json
import numpy as np
from hestia_earth.schema import EmissionMethodTier


EXCLUDE_FIELDS = ["@type", "type", "@context"]
EXCLUDE_PRIVATE_FIELDS = [
    "added",
    "addedVersion",
    "updated",
    "updatedVersion",
    "aggregatedVersion",
    "_cache",
]


def _with_csv_formatting(dct):
    """
    Use as object_hook when parsing a JSON node: json.loads(node, object_hook=_with_csv_formatting).
    Ensures parsed JSON has field values formatted according to hestia csv conventions.
    """
    if "boundary" in dct:
        dct["boundary"] = json.dumps(dct["boundary"])
    for key, value in dct.items():
        if _is_scalar_list(value):
            dct[key] = ";".join([str(el) for el in value])
    return dct


def _is_scalar_list(value):
    if not isinstance(value, list):
        return False
    all_scalar = True
    for element in value:
        if not np.isscalar(element):
            all_scalar = False
            break
    return all_scalar


def _filter_not_relevant(blank_node: dict):
    return blank_node.get("methodTier") != EmissionMethodTier.NOT_RELEVANT.value


def _filter_emissions_not_relevant(node: dict):
    """
    Ignore all emissions where `methodTier=not relevant` so save space.
    """
    return node | (
        {
            key: list(filter(_filter_not_relevant, node[key]))
            for key in ["emissions", "emissionsResourceUse"]
            if key in node
        }
    )



import os


def _get_folder(glossary: bool = False) -> str:
    return (
        os.getenv("DOWNLOAD_FOLDER_GLOSSARY")
        if glossary
        else os.getenv("DOWNLOAD_FOLDER")
    )


def _load_from_folder(folder: str, key: str):
    try:
        with open(os.path.join(folder, key)) as f:
            return f.read().encode("utf-8")
    except Exception:
        # in case the file does not exist, should simply return None
        return None


def _exists_in_folder(folder: str, key: str):
    return os.path.exists(os.path.join(folder, key))
